
from __future__ import annotations

from collections import OrderedDict
from typing import Optional, Dict, List, Callable, Union, Tuple, Awaitable, \
    Iterable, NoReturn

from pymap.bytes import MaybeBytes
from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.context import socket_info, connection_exit
from pymap.exceptions import NotSupportedError, CloseConnection
from pymap.fetch import MessageAttributes
from pymap.interfaces.login import LoginInterface
from pymap.interfaces.session import SessionInterface
from pymap.parsing.command import CommandAuth, CommandNonAuth, CommandSelect, \
    Command
from pymap.parsing.command.any import CapabilityCommand, LogoutCommand, \
    NoOpCommand
from pymap.parsing.command.nonauth import AuthenticateCommand, LoginCommand, \
    StartTLSCommand
from pymap.parsing.command.auth import AppendCommand, CreateCommand, \
    DeleteCommand, ListCommand, RenameCommand, SelectCommand, StatusCommand, \
    SubscribeCommand, UnsubscribeCommand
from pymap.parsing.command.select import CheckCommand, CloseCommand, \
    IdleCommand, ExpungeCommand, CopyCommand, MoveCommand, FetchCommand, \
    StoreCommand, SearchCommand
from pymap.parsing.commands import InvalidCommand
from pymap.parsing.primitives import ListP, Number
from pymap.parsing.response import ResponseOk, ResponseNo, ResponseBad, \
    ResponseCode, ResponsePreAuth, CommandResponse, UntaggedResponse
from pymap.parsing.response.code import Capability, PermanentFlags, UidNext, \
    UidValidity, Unseen, MailboxId
from pymap.parsing.response.specials import FlagsResponse, ExistsResponse, \
    RecentResponse, FetchResponse, ListResponse, LSubResponse, \
    SearchResponse, StatusResponse
from pymap.parsing.specials import StatusAttribute, FetchAttribute, FetchValue
from pymap.selected import SelectedMailbox
from pysasl import AuthenticationCredentials

__all__ = ['ConnectionState']

_AuthCommands = Union[AuthenticateCommand, LoginCommand]
_CommandRet = Tuple[CommandResponse, Optional[SelectedMailbox]]
_CommandFunc = Callable[[Command], Awaitable[_CommandRet]]

_flags_attr = FetchAttribute(b'FLAGS')
_uid_attr = FetchAttribute(b'UID')


class ConnectionState:
    """Defines the state flow of the IMAP connection. Determines if a command
    can be processed at that point in the connection, and interacts with the
    backend plugin.

    """

    def __init__(self, login: LoginInterface, config: IMAPConfig) -> None:
        super().__init__()
        self.config = config
        self.auth = config.initial_auth
        self.login = login
        self._session: Optional[SessionInterface] = None
        self._selected: Optional[SelectedMailbox] = None
        self._capability = list(config.initial_capability)

    @property
    def session(self) -> SessionInterface:
        if self._session is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._session

    @property
    def selected(self) -> SelectedMailbox:
        if self._selected is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._selected

    @property
    def capability(self) -> Capability:
        if self._session:
            return Capability(self._capability)
        else:
            if self.auth.get_server(b'PLAIN') is None:
                logindisabled = [b'LOGINDISABLED']
            else:
                logindisabled = []
            return Capability(self._capability + logindisabled +
                              [b'AUTH=%b' % mech.name for mech in
                               self.auth.server_mechanisms])

    async def do_cleanup(self) -> None:
        try:
            await self.session.cleanup()
        except Exception:
            pass

    async def _login(self, creds: AuthenticationCredentials) \
            -> SessionInterface:
        stack = connection_exit.get()
        identity = await self.login.authenticate(creds)
        return await stack.enter_async_context(identity.new_session())

    async def do_greeting(self) -> CommandResponse:
        preauth_creds = self.config.preauth_credentials
        if preauth_creds:
            self._session = await self._login(preauth_creds)
        elif socket_info.get().from_localhost:
            self.auth = self.config.tls_auth
        resp_cls = ResponsePreAuth if preauth_creds else ResponseOk
        return resp_cls(b'*', self.config.greeting, self.capability)

    async def do_authenticate(self, cmd: _AuthCommands,
                              creds: Optional[AuthenticationCredentials]) \
            -> CommandResponse:
        if not creds:
            return ResponseNo(cmd.tag, b'Invalid authentication mechanism.')
        self._session = await self._login(creds)
        self._capability.extend(self.config.login_capability)
        return ResponseOk(cmd.tag, b'Authentication successful.',
                          self.capability)

    async def do_login(self, cmd: LoginCommand) -> _CommandRet:
        if b'LOGINDISABLED' in self.capability:
            raise NotSupportedError('LOGIN is disabled.')
        creds = AuthenticationCredentials(
            cmd.userid.decode('utf-8', 'surrogateescape'),
            cmd.password.decode('utf-8', 'surrogateescape'))
        return await self.do_authenticate(cmd, creds), None

    async def do_starttls(self, cmd: StartTLSCommand) -> _CommandRet:
        try:
            self._capability.remove(b'STARTTLS')
        except ValueError:
            raise NotSupportedError('STARTTLS not available.')
        self.auth = self.config.tls_auth
        return ResponseOk(cmd.tag, b'Ready to handshake.'), None

    async def do_capability(self, cmd: CapabilityCommand) -> _CommandRet:
        response = ResponseOk(cmd.tag, b'Capabilities listed.')
        response.add_untagged(UntaggedResponse(self.capability.string))
        return response, None

    async def do_noop(self, cmd: NoOpCommand) -> _CommandRet:
        updates = None
        if self._selected and self._session:
            updates = await self.session.check_mailbox(self.selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_select(self, cmd: SelectCommand) -> _CommandRet:
        self._selected = None
        mailbox, updates = await self.session.select_mailbox(
            cmd.mailbox, cmd.readonly)
        if updates.readonly:
            num_recent = mailbox.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-ONLY'))
            resp.add_untagged_ok(b'Read-only mailbox.', PermanentFlags([]))
        else:
            num_recent = updates.session_flags.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-WRITE'))
            resp.add_untagged_ok(b'Flags permitted.',
                                 PermanentFlags(mailbox.permanent_flags))
        messages = updates.messages
        resp.add_untagged(FlagsResponse(mailbox.flags))
        resp.add_untagged(ExistsResponse(messages.exists))
        resp.add_untagged(RecentResponse(num_recent))
        resp.add_untagged_ok(b'Predicted next UID.',
                             UidNext(mailbox.next_uid))
        resp.add_untagged_ok(b'UIDs valid.',
                             UidValidity(mailbox.uid_validity))
        if mailbox.first_unseen:
            resp.add_untagged_ok(b'First unseen message.',
                                 Unseen(mailbox.first_unseen))
        resp.add_untagged_ok(b'Object ID.', MailboxId(mailbox.mailbox_id))
        return resp, updates

    async def do_create(self, cmd: CreateCommand) -> _CommandRet:
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot create INBOX.'), None
        mailbox_id, updates = await self.session.create_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.',
                          MailboxId(mailbox_id)), updates

    async def do_delete(self, cmd: DeleteCommand) -> _CommandRet:
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot delete INBOX.'), None
        updates = await self.session.delete_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_rename(self, cmd: RenameCommand) -> _CommandRet:
        if cmd.to_mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot rename to INBOX.'), None
        updates = await self.session.rename_mailbox(
            cmd.from_mailbox, cmd.to_mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_status(self, cmd: StatusCommand) -> _CommandRet:
        mailbox, updates = await self.session.get_mailbox(
            cmd.mailbox, selected=self._selected)
        data: Dict[StatusAttribute, MaybeBytes] = OrderedDict()
        for attr in cmd.status_list:
            if attr == b'MESSAGES':
                data[attr] = Number(mailbox.exists)
            elif attr == b'RECENT':
                if updates and updates.mailbox_id == mailbox.mailbox_id:
                    data[attr] = Number(updates.session_flags.recent)
                else:
                    data[attr] = Number(mailbox.recent)
            elif attr == b'UNSEEN':
                data[attr] = Number(mailbox.unseen)
            elif attr == b'UIDNEXT':
                data[attr] = Number(mailbox.next_uid)
            elif attr == b'UIDVALIDITY':
                data[attr] = Number(mailbox.uid_validity)
            elif attr == b'MAILBOXID':
                data[attr] = mailbox.mailbox_id.parens
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp.add_untagged(StatusResponse(cmd.mailbox, data))
        return resp, updates

    async def do_append(self, cmd: AppendCommand) -> _CommandRet:
        if len(cmd.messages) > 1 and b'MULTIAPPEND' not in self.capability:
            raise NotSupportedError('MULTIAPPEND is disabled.')
        if cmd.cancelled:
            return ResponseNo(cmd.tag, b'APPEND cancelled.'), None
        if cmd.error:
            raise cmd.error
        append_uid, updates = await self.session.append_messages(
            cmd.mailbox, cmd.messages, selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', append_uid)
        return resp, updates

    async def do_subscribe(self, cmd: SubscribeCommand) -> _CommandRet:
        updates = await self.session.subscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_unsubscribe(self, cmd: UnsubscribeCommand) -> _CommandRet:
        updates = await self.session.unsubscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_list(self, cmd: ListCommand) -> _CommandRet:
        mailboxes, updates = await self.session.list_mailboxes(
            cmd.ref_name, cmd.filter, subscribed=cmd.only_subscribed,
            selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp_type = LSubResponse if cmd.only_subscribed else ListResponse
        for name, sep, attrs in mailboxes:
            resp.add_untagged(resp_type(name, sep, attrs))
        return resp, updates

    async def do_check(self, cmd: CheckCommand) -> _CommandRet:
        updates = await self.session.check_mailbox(
            self.selected, housekeeping=True)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_close(self, cmd: CloseCommand) -> _CommandRet:
        await self.session.expunge_mailbox(self.selected)
        self._selected = None
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    async def do_expunge(self, cmd: ExpungeCommand) -> _CommandRet:
        updates = await self.session.expunge_mailbox(
            self.selected, cmd.uid_set)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        return resp, updates

    async def do_copy(self, cmd: CopyCommand) -> _CommandRet:
        copy_uid, updates = await self.session.copy_messages(
            self.selected, cmd.sequence_set, cmd.mailbox)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', copy_uid)
        return resp, updates

    async def do_move(self, cmd: MoveCommand) -> _CommandRet:
        copy_uid, updates = await self.session.move_messages(
            self.selected, cmd.sequence_set, cmd.mailbox)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp.add_untagged_ok(b'Moved.', copy_uid)
        return resp, updates

    async def do_fetch(self, cmd: FetchCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        set_seen = not self.selected.readonly and \
            any(attr.set_seen for attr in cmd.attributes)
        messages, updates = await self.session.fetch_messages(
            self.selected, cmd.sequence_set, set_seen)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            msg_attrs = MessageAttributes(msg, self.selected, cmd.attributes)
            fetch_resp = FetchResponse(msg_seq, msg_attrs,
                                       writing_hook=msg_attrs.load_hook())
            resp.add_untagged(fetch_resp)
        return resp, updates

    async def do_search(self, cmd: SearchCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        messages, updates = await self.session.search_mailbox(
            self.selected, cmd.keys)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        msg_ids: List[int] = []
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            if cmd.uid:
                msg_ids.append(msg.uid)
            else:
                msg_ids.append(msg_seq)
        resp.add_untagged(SearchResponse(msg_ids))
        return resp, updates

    async def do_store(self, cmd: StoreCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        if cmd.silent:
            self.selected.silence(cmd.sequence_set, cmd.flag_set, cmd.mode)
        messages, updates = await self.session.update_flags(
            self.selected, cmd.sequence_set, cmd.flag_set, cmd.mode)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        session_flags = self.selected.session_flags
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            elif cmd.silent:
                continue
            flags = msg.get_flags(session_flags)
            fetch_data: List[FetchValue] = [
                FetchValue.of(_flags_attr, ListP(flags, sort=True))]
            if cmd.uid:
                fetch_data.append(
                    FetchValue.of(_uid_attr, Number(msg.uid)))
            resp.add_untagged(FetchResponse(msg_seq, fetch_data))
        return resp, updates

    async def do_idle(self, cmd: IdleCommand) -> _CommandRet:
        if b'IDLE' not in self.capability:
            raise NotSupportedError('IDLE is disabled.')
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    @classmethod
    async def do_logout(cls, cmd: LogoutCommand) -> NoReturn:
        raise CloseConnection()

    async def receive_updates(self, cmd: IdleCommand, done: Event) \
            -> Iterable[UntaggedResponse]:
        selected = await self.session.check_mailbox(
            self.selected, wait_on=done)
        self._selected, untagged = selected.fork(cmd)
        return untagged

    @classmethod
    def _get_func_name(cls, cmd: Command) -> str:
        cmd_type = type(cmd)
        while cmd_type.delegate:
            cmd_type = cmd_type.delegate
        cmd_str = str(cmd_type.command, 'ascii').lower()
        return 'do_' + cmd_str

    async def do_command(self, cmd: Command) -> CommandResponse:
        if isinstance(cmd, InvalidCommand):
            return ResponseBad(cmd.tag, cmd.message)
        elif self._session and isinstance(cmd, CommandNonAuth):
            msg = cmd.command + b': Already authenticated.'
            return ResponseBad(cmd.tag, msg)
        elif not self._session and isinstance(cmd, CommandAuth):
            msg = cmd.command + b': Must authenticate first.'
            return ResponseBad(cmd.tag, msg)
        elif not self._selected and isinstance(cmd, CommandSelect):
            msg = cmd.command + b': Must select a mailbox first.'
            return ResponseBad(cmd.tag, msg)
        func_name = self._get_func_name(cmd)
        try:
            func: _CommandFunc = getattr(self, func_name)
        except AttributeError:
            return ResponseNo(cmd.tag, cmd.command + b': Not Implemented')
        response, selected = await func(cmd)
        if selected is not None:
            self._selected, untagged = selected.fork(cmd)
            response.add_untagged(*untagged)
        return response




from __future__ import annotations

from itertools import chain
from typing import Iterable, Iterator, Tuple, Union, Sequence, Optional, \
    List, FrozenSet

from .. import Params, Parseable, Space
from ..exceptions import NotParseable
from ...bytes import rev

__all__ = ['MaxValue', 'SequenceSet']

_SeqIdx = Union['MaxValue', int]
_SeqElem = Union[_SeqIdx, Tuple[_SeqIdx, _SeqIdx]]


class MaxValue:
    """The type used as a placeholder for the maximum value."""

    def __eq__(self, other) -> bool:
        if isinstance(other, MaxValue):
            return True
        return NotImplemented

    def __hash__(self) -> int:
        return hash(type(self))


class SequenceSet(Parseable[Sequence[_SeqElem]]):
    """Represents a sequence set from an IMAP stream.

    Args:
        sequences: The sequence set data.
        uid: True if the sequences refer to message UIDs.

    Attributes:
        uid: True if the sequences refer to message UIDs.

    """

    _num_pattern = rev.compile(br'[1-9]\d*')
    _max = MaxValue()

    def __init__(self, sequences: Sequence[_SeqElem],
                 uid: bool = False) -> None:
        super().__init__()
        self.sequences = sequences
        self.uid = uid
        self._raw: Optional[bytes] = None

    @classmethod
    def all(cls, uid: bool = False) -> SequenceSet:
        """A sequence set intended to contain all values."""
        return _AllSequenceSet(uid)

    @property
    def value(self) -> Sequence[_SeqElem]:
        """The sequence set data."""
        return self.sequences

    @property
    def is_all(self) -> bool:
        """True if the sequence set starts at ``1`` and ends at the maximum
        value.

        This may be used to optimize cases of checking for a value in the set,
        avoiding the need to provide ``max_value`` in :meth:`.flatten` or
        :meth:`.iter`.

        """
        first = self.sequences[0]
        return isinstance(first, tuple) \
            and first[0] == 1 and isinstance(first[1], MaxValue)

    @classmethod
    def _get_range(cls, elem: _SeqElem, max_value: int) -> Iterable[int]:
        if isinstance(elem, int):
            if elem <= max_value:
                return range(elem, elem + 1)
            else:
                return ()
        elif isinstance(elem, MaxValue):
            return range(max_value, max_value + 1)
        else:
            left, right = elem
            if isinstance(left, MaxValue):
                left = max_value
            if isinstance(right, MaxValue):
                right = max_value
            low = min(left, right)
            if low <= max_value:
                high = min(max(left, right), max_value)
                return range(low, high + 1)
            else:
                return ()

    def flatten(self, max_value: int) -> FrozenSet[int]:
        """Return a set of all values contained in the sequence set.

        Args:
            max_value: The maximum value, in place of any ``*``.

        """
        return frozenset(self.iter(max_value))

    def iter(self, max_value: int) -> Iterator[int]:
        """Iterates through the sequence numbers contained in the set, bounded
        by the given maximum value (in place of any ``*``).

        Args:
            max_value: The maximum value of the set.

        """
        return chain.from_iterable(
            (self._get_range(elem, max_value) for elem in self.sequences))

    def _elem_bytes(self, elem: _SeqIdx) -> bytes:
        if isinstance(elem, MaxValue):
            return b'*'
        else:
            return b'%d' % elem

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        parts = []
        for group in self.value:
            if isinstance(group, tuple):
                left = self._elem_bytes(group[0])
                right = self._elem_bytes(group[1])
                parts.append(b'%b:%b' % (left, right))
            else:
                parts.append(self._elem_bytes(group))
        self._raw = raw = b','.join(parts)
        return raw

    def __eq__(self, other) -> bool:
        if isinstance(other, SequenceSet):
            return self.uid == other.uid \
                and self.sequences == other.sequences
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash((type(self), tuple(self.sequences), self.uid))

    def __repr__(self) -> str:
        attr = 'uidset' if self.uid else 'set'
        return '<SequenceSet %s=%r>' % (attr, self.sequences)

    @classmethod
    def _parse_part(cls, buf: memoryview) -> Tuple[_SeqElem, memoryview]:
        if buf and buf[0] == 0x2a:
            item1: _SeqIdx = cls._max
            buf = buf[1:]
        else:
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                item1 = int(match.group(0))
            else:
                raise NotParseable(buf)
        if buf and buf[0] == 0x3a:
            buf = buf[1:]
            if buf and buf[0] == 0x2a:
                return (item1, cls._max), buf[1:]
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                return (item1, int(match.group(0))), buf
            raise NotParseable(buf)
        return item1, buf

    @classmethod
    def build(cls, seqs: Iterable[int], uid: bool = False) -> SequenceSet:
        """Build a new sequence set that contains the given values using as
        few groups as possible.

        Args:
            seqs: The sequence values to build.
            uid: True if the sequences refer to message UIDs.

        """
        seqs_list = sorted(set(seqs))
        groups: List[Union[int, Tuple[int, int]]] = []
        group: Union[int, Tuple[int, int]] = seqs_list[0]
        for i in range(1, len(seqs_list)):
            group_i = seqs_list[i]
            if isinstance(group, int):
                if group_i == group + 1:
                    group = (group, group_i)
                else:
                    groups.append(group)
                    group = group_i
            elif isinstance(group, tuple):
                if group_i == group[1] + 1:
                    group = (group[0], group_i)
                else:
                    groups.append(group)
                    group = group_i
        groups.append(group)
        return SequenceSet(groups, uid)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[SequenceSet, memoryview]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        sequences = []
        while buf:
            item, buf = cls._parse_part(buf)
            sequences.append(item)
            if buf and buf[0] != 0x2c:
                break
            buf = buf[1:]
        if not sequences:
            raise NotParseable(buf)
        return cls(sequences, uid=params.uid), buf


class _AllSequenceSet(SequenceSet):

    def __init__(self, uid: bool) -> None:
        super().__init__([(1, self._max)], uid)

    def iter(self, max_value: int) -> Iterator[int]:
        return iter(range(1, max_value + 1))

    def __bytes__(self) -> bytes:
        return b'1:*'

    def __repr__(self) -> str:
        return '<SequenceSet set=all>'




from __future__ import annotations

import asyncio
import sys
from argparse import ArgumentParser, SUPPRESS
from asyncio import CancelledError
from datetime import datetime, timedelta, timezone
from secrets import token_bytes
from ssl import SSLContext
from typing import Optional, Sequence, Awaitable

from grpclib.events import listen, RecvRequest
from grpclib.server import Server
from pymap.interfaces.backend import BackendInterface, ServiceInterface
from pymapadmin import is_compatible, __version__ as server_version
from pymapadmin.local import get_admin_socket

from .errors import get_incompatible_version_error
from .handlers import handlers
from .token import get_admin_token

__all__ = ['AdminService']


class AdminService(ServiceInterface):  # pragma: no cover
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    This service can be exposed in *private* or *public* mode. Private mode
    does not use SSL/TLS encryption and authentication is not required, so it
    should be appropriately firewalled. Public mode requires both SSL/TLS
    encryption and the client provide credentials for the user being
    operated on, e.g. a user can change their own password but cannot change
    another user's password.

    The `pymap-admin <https://github.com/icgood/pymap-admin>`_ CLI may be used
    to interact with the admin service and serves as a reference client
    implementation.

    """

    _admin_token: Optional[bytes] = None

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-host', action='append', metavar='IFACE',
                           help='the network interface to listen on')
        group.add_argument('--admin-port', metavar='NUM',
                           help='the port or service name to listen on')
        group.add_argument('--no-admin-token', action='store_true',
                           help='disable the admin token')
        group.add_argument('--admin-token-duration', metavar='SEC', type=int,
                           help=SUPPRESS)

    def _init_admin_token(self) -> None:
        if not self.config.args.no_admin_token:
            expiration: Optional[datetime] = None
            if self.config.args.admin_token_duration:
                duration_sec: int = self.config.args.admin_token_duration
                duration = timedelta(seconds=duration_sec)
                expiration = datetime.now(tz=timezone.utc) + duration
            self._admin_token = token_bytes()
            token = get_admin_token(self._admin_token, expiration)
            print(f'PYMAP_ADMIN_TOKEN={token.serialize()}', file=sys.stderr)

    async def _check_version(self, event: RecvRequest) -> None:
        client_version = event.metadata['client-version']
        if isinstance(client_version, bytes):
            client_version = client_version.decode('ascii')
        if not is_compatible(client_version, server_version):
            raise get_incompatible_version_error(
                client_version, server_version)

    async def start(self) -> Awaitable:
        self._init_admin_token()
        backend = self.backend
        host: Optional[str] = self.config.args.admin_host
        port: Optional[int] = self.config.args.admin_port
        ssl = self.config.ssl_context
        servers = [await self._start_local(backend)]
        if host or port:
            servers += [await self._start(backend, host, port, ssl)]
        return asyncio.create_task(self._run(servers))

    def _new_server(self, backend: BackendInterface) -> Server:
        server = Server([handler(backend, self._admin_token)
                         for _, handler in handlers])
        listen(server, RecvRequest, self._check_version)
        return server

    async def _start_local(self, backend: BackendInterface) -> Server:
        server = self._new_server(backend)
        path = get_admin_socket(mkdir=True)
        await server.start(path=path)
        return server

    async def _start(self, backend: BackendInterface, host: Optional[str],
                     port: Optional[int], ssl: Optional[SSLContext]) -> Server:
        server = self._new_server(backend)
        await server.start(host=host, port=port, ssl=ssl, reuse_address=True)
        return server

    async def _run(self, servers: Sequence[Server]) -> None:
        try:
            for server in servers:
                await server.wait_closed()
        except CancelledError:
            for server in servers:
                server.close()
                await server.wait_closed()




from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import Optional
from typing_extensions import Final

from pymacaroons import Macaroon, Verifier
from pymacaroons.exceptions import MacaroonDeserializationException, \
    MacaroonInvalidSignatureException
from pysasl.creds import StoredSecret, AuthenticationCredentials

__all__ = ['get_login_token', 'get_admin_token', 'TokenCredentials']


def get_login_token(identifier: str, key: bytes, *, location: str = None,
                    expiration: datetime = None,
                    authzid: str = None) -> Macaroon:
    """Produce a token :class:`~pymacaroons.Macaroon` to authenticate as a
    specific user.

    Args:
        identifier: Used to identify and lookup the *key* value.
        key: The key used to create and verify the token for the user.
        location: Optional hint describing where the token is valid.
        expiration: When the token should expire.
        authzid: If given, limits the token to the given authorization ID.

    """
    macaroon = Macaroon(location=location, identifier=identifier, key=key)
    macaroon.add_first_party_caveat('type = login')
    if authzid is not None:
        macaroon.add_first_party_caveat(f'authzid = {authzid}')
    if expiration is not None:
        macaroon.add_first_party_caveat(f'time < {expiration.isoformat()}')
    return macaroon


def get_admin_token(admin_token: bytes, expiration: Optional[datetime], *,
                    location: str = None, authzid: str = None) -> Macaroon:
    """Produce a token :class:`~pymacaroons.Macaroon` to authenticate as an
    admin that may perform operations on users.

    Args:
        admin_token: The admin token string.
        expiration: When the token should expire.
        location: Optional hint describing where the token is valid.
        authzid: If given, limits the token to the given authorization ID.

    """
    macaroon = Macaroon(location=location, identifier='', key=admin_token)
    macaroon.add_first_party_caveat('type = admin')
    if authzid is not None:
        macaroon.add_first_party_caveat(f'authzid = {authzid}')
    if expiration is not None:
        macaroon.add_first_party_caveat(f'time < {expiration.isoformat()}')
    return macaroon


class TokenCredentials(AuthenticationCredentials):
    """Authenticate using a `Macaroon`_ token. Tokens may be created with
    either :func:`get_login_token` or :func:`get_admin_token`.

    Either type of token may use a caveat with ``time < `` and a
    :func:`~datetime.datetime.fromisoformat` string to limit how long the token
    is valid.

    .. Macaroon: https://github.com/ecordell/pymacaroons#readme

    Args:
        serialized: The serialized macaroon string.
        admin_token: The admin token string.
        identity: The identity to be authorized.

    """

    _caveat = re.compile(r'^(\w+) ([^\w\s]+) (.*)$', re.ASCII)

    def __init__(self, serialized: str, admin_token: Optional[bytes],
                 identity: str) -> None:
        try:
            macaroon = Macaroon.deserialize(serialized)
        except MacaroonDeserializationException as exc:
            raise ValueError('invalid token') from exc
        super().__init__(macaroon.identifier, '', identity)
        self.macaroon: Final = macaroon
        self.admin_token: Final = admin_token
        self._type = self._find_type(macaroon)

    @classmethod
    def _find_type(cls, macaroon: Macaroon) -> Optional[str]:
        for caveat in macaroon.first_party_caveats():
            caveat_id = caveat.caveat_id
            if caveat_id == 'type = admin':
                return 'admin-token'
            elif caveat_id == 'type = login':
                return 'login-token'
        return None

    @property
    def authcid_type(self) -> Optional[str]:
        return self._type

    @property
    def has_secret(self) -> bool:
        return False

    @property
    def secret(self) -> str:
        raise AttributeError('secret')

    def _satisfy(self, predicate: str) -> bool:
        match = self._caveat.match(predicate)
        if match is None:
            return False
        key, op, value = match.groups()
        if key == 'time' and op == '<':
            try:
                end = datetime.fromisoformat(value)
            except ValueError:
                return False
            now = datetime.now(tz=timezone.utc)
            return now < end
        elif key == 'authzid' and op == '=':
            return value == self.authzid
        else:
            return False

    def _verify(self, verifier: Verifier, key: bytes) -> bool:
        try:
            return verifier.verify(self.macaroon, key)
        except MacaroonInvalidSignatureException:
            return False

    def _get_login_verifier(self) -> Verifier:
        verifier = Verifier()
        verifier.satisfy_general(self._satisfy)
        verifier.satisfy_exact('type = login')
        return verifier

    def _get_admin_verifier(self) -> Verifier:
        verifier = Verifier()
        verifier.satisfy_general(self._satisfy)
        verifier.satisfy_exact('type = admin')
        return verifier

    def check_secret(self, secret: Optional[StoredSecret], *,
                     key: bytes = None, **other: Optional[str]) -> bool:
        if key is not None:
            verifier = self._get_login_verifier()
            if self._verify(verifier, key):
                return True
        admin_token = self.admin_token
        if admin_token is not None:
            verifier = self._get_admin_verifier()
            if self._verify(verifier, admin_token):
                return True
        return False




from __future__ import annotations

from typing import ClassVar, Type, Optional, Tuple, Mapping, Sequence
from typing_extensions import Final

from pymap.parsing import Parseable, Params, EndLine
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import Atom, Number, String, QuotedString

__all__ = ['Command', 'NoOpCommand', 'CapabilityCommand', 'StartTLSCommand',
           'AuthenticateCommand', 'UnauthenticateCommand', 'LogoutCommand',
           'HaveSpaceCommand', 'PutScriptCommand', 'ListScriptsCommand',
           'SetActiveCommand', 'GetScriptCommand', 'DeleteScriptCommand',
           'RenameScriptCommand', 'CheckScriptCommand']


class Command(Parseable[bytes]):
    """Base class for ManageSieve commands."""

    #: The command key, e.g. ``b'NOOP'``.
    command: ClassVar[bytes] = b''

    _commands: Optional[Mapping[bytes, Type[Command]]] = None

    @property
    def value(self) -> bytes:
        """The command name."""
        return self.command

    def __bytes__(self) -> bytes:
        return self.command

    def __repr__(self) -> str:
        return '<%s>' % (type(self).__name__)

    @classmethod
    def _all_commands(cls) -> Sequence[Type[Command]]:
        return [
            NoOpCommand, CapabilityCommand, StartTLSCommand,
            AuthenticateCommand, UnauthenticateCommand, LogoutCommand,
            HaveSpaceCommand, PutScriptCommand, ListScriptsCommand,
            SetActiveCommand, GetScriptCommand, DeleteScriptCommand,
            RenameScriptCommand, CheckScriptCommand]

    @classmethod
    def _get_commands(cls) -> Mapping[bytes, Type[Command]]:
        if cls._commands is None:
            mapping = {}
            for cmd in cls._all_commands():
                mapping[cmd.command] = cmd
            cls._commands = mapping
        return cls._commands

    @classmethod
    def _parse_script_name(self, buf: memoryview, params: Params,
                           allow_empty: bool = False) \
            -> Tuple[str, memoryview]:
        name_obj, after = String.parse(buf, params)
        if not allow_empty and not name_obj.value:
            raise NotParseable(buf)
        try:
            return name_obj.value.decode('utf-8'), after
        except UnicodeError as exc:
            raise NotParseable(buf) from exc

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[Command, memoryview]:
        commands = cls._get_commands()
        cmd_name, after = Atom.parse(buf, params)
        try:
            cmd_type = commands[cmd_name.value.upper()]
        except KeyError:
            raise NotParseable(buf)
        else:
            buf = after
        cmd, buf = cmd_type.parse(buf, params)
        return cmd, buf


class NoOpCommand(Command):

    command = b'NOOP'

    def __init__(self, tag: Optional[bytes]) -> None:
        super().__init__()
        self.tag: Final = tag

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[NoOpCommand, memoryview]:
        whitespace = cls._whitespace_length(buf)
        tag: Optional[bytes] = None
        if whitespace > 0:
            buf = buf[whitespace:]
            try:
                tag_obj, buf = String.parse(buf, params)
            except NotParseable:
                pass
            else:
                tag = tag_obj.value
        _, buf = EndLine.parse(buf, params)
        return cls(tag), buf


class CapabilityCommand(Command):

    command = b'CAPABILITY'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[CapabilityCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class StartTLSCommand(Command):

    command = b'STARTTLS'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[StartTLSCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class AuthenticateCommand(Command):

    command = b'AUTHENTICATE'

    def __init__(self, mech_name: bytes, initial_data: bytes = None) -> None:
        super().__init__()
        self.mech_name: Final = mech_name
        self.initial_data: Final = initial_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[AuthenticateCommand, memoryview]:
        mech_name, buf = QuotedString.parse(buf, params)
        try:
            data_obj, buf = String.parse(buf, params)
        except NotParseable:
            initial_data: Optional[bytes] = None
        else:
            initial_data = data_obj.value
        return cls(mech_name.value, initial_data), buf


class UnauthenticateCommand(Command):

    command = b'UNAUTHENTICATE'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[UnauthenticateCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class LogoutCommand(Command):

    command = b'LOGOUT'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[LogoutCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class HaveSpaceCommand(Command):

    command = b'HAVESPACE'

    def __init__(self, script_name: str, size: int) -> None:
        super().__init__()
        self.script_name: Final = script_name
        self.size: Final = size

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[HaveSpaceCommand, memoryview]:
        script_name, after = cls._parse_script_name(buf, params)
        if not script_name:
            raise NotParseable(buf)
        size, buf = Number.parse(after, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name, size.value), buf


class PutScriptCommand(Command):

    command = b'PUTSCRIPT'

    def __init__(self, script_name: str, script_data: bytes) -> None:
        super().__init__()
        self.script_name: Final = script_name
        self.script_data: Final = script_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[PutScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        script_data, buf = String.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name, script_data.value), buf


class ListScriptsCommand(Command):

    command = b'LISTSCRIPTS'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[ListScriptsCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class SetActiveCommand(Command):

    command = b'SETACTIVE'

    def __init__(self, script_name: Optional[str]) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[SetActiveCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params, True)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name or None), buf


class GetScriptCommand(Command):

    command = b'GETSCRIPT'

    def __init__(self, script_name: str) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[GetScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name), buf


class DeleteScriptCommand(Command):

    command = b'DELETESCRIPT'

    def __init__(self, script_name: str) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[DeleteScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name), buf


class RenameScriptCommand(Command):

    command = b'RENAMESCRIPT'

    def __init__(self, old_script_name: str, new_script_name: str) -> None:
        super().__init__()
        self.old_script_name: Final = old_script_name
        self.new_script_name: Final = new_script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[RenameScriptCommand, memoryview]:
        old_script_name, buf = cls._parse_script_name(buf, params)
        new_script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(old_script_name, new_script_name), buf


class CheckScriptCommand(Command):

    command = b'CHECKSCRIPT'

    def __init__(self, script_data: bytes) -> None:
        super().__init__()
        self.script_data: Final = script_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[CheckScriptCommand, memoryview]:
        script_data, buf = String.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_data.value), buf




from __future__ import annotations

from typing import TypeVar, Generic, Callable, Iterable, Iterator, \
    Tuple, Mapping, Dict
from typing_extensions import Final

from pkg_resources import iter_entry_points, DistributionNotFound

__all__ = ['PluginT', 'Plugin']

#: The plugin type variable.
PluginT = TypeVar('PluginT')


class Plugin(Generic[PluginT], Iterable[Tuple[str, PluginT]]):
    """Plugin system, typically loaded from :mod:`pkg_resources` `entry points
    <https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata>`_.

    >>> example: Plugin[Type[Example]] = Plugin('plugins.example')
    >>> example.add('two', ExampleTwo)
    >>> example.registered
    {'one': ExampleOne, 'two': ExampleTwo}

    Note:
        Plugins registered from *group* entry points are lazy-loaded. This is
        to avoid cyclic imports.

    Args:
        group: The entry point group to load.

    """

    def __init__(self, group: str) -> None:
        super().__init__()
        self.group: Final = group
        self._loaded = False
        self._registered: Dict[str, PluginT] = {}

    def __iter__(self) -> Iterator[Tuple[str, PluginT]]:
        return iter(self.registered.items())

    @property
    def registered(self) -> Mapping[str, PluginT]:
        """A mapping of the registered plugins, keyed by name."""
        self._load()
        return self._registered

    @property
    def first(self) -> PluginT:
        """The first registered plugin."""
        first = next(iter(self.registered.values()), None)
        assert first is not None, \
            f'plugin group {self.group} has no entries'
        return first

    def _load(self) -> None:
        if not self._loaded:
            for entry_point in iter_entry_points(self.group):
                try:
                    plugin: PluginT = entry_point.load()
                except DistributionNotFound:
                    pass  # optional dependencies not installed
                else:
                    self.add(entry_point.name, plugin)
            self._loaded = True

    def add(self, name: str, plugin: PluginT) -> None:
        """Add a new plugin by name.

        Args:
            name: The identifying name of the plugin.
            plugin: The plugin object.

        """
        assert name not in self._registered, \
            f'plugin {self.group}:{name} has already been registered'
        self._registered[name] = plugin

    def register(self, name: str) -> Callable[[PluginT], PluginT]:
        """Decorates a plugin implementation.

        Args:
            name: The identifying name of the plugin.

        """
        def deco(plugin: PluginT) -> PluginT:
            self.add(name, plugin)
            return plugin
        return deco

    def __repr__(self) -> str:
        return f'Plugin({self.group!r})'




from __future__ import annotations

from functools import total_ordering
from typing import Tuple, FrozenSet, Union

from .. import Params, Parseable, Space
from ..exceptions import NotParseable
from ..primitives import Atom

__all__ = ['Flag', 'get_system_flags', 'Seen', 'Recent', 'Deleted', 'Flagged',
           'Answered', 'Draft', 'Wildcard']


@total_ordering
class Flag(Parseable[bytes]):
    """Represents a message flag from an IMAP stream.

    Args:
        value: The flag or keyword value.

    """

    def __init__(self, value: Union[str, bytes]) -> None:
        super().__init__()
        if isinstance(value, bytes):
            value_bytes = value
        else:
            value_bytes = bytes(value, 'ascii')
        self._value = self._capitalize(value_bytes)
        self._hash = hash(self._value)

    @property
    def value(self) -> bytes:
        """The flag or keyword value."""
        return self._value

    @property
    def is_system(self) -> bool:
        """True if the flag is an RFC-defined IMAP system flag."""
        return self.value.startswith(b'\\')

    @classmethod
    def _capitalize(cls, value: bytes) -> bytes:
        if value.startswith(b'\\'):
            return b'\\' + value[1:].capitalize()
        return value

    def __eq__(self, other) -> bool:
        if isinstance(other, Flag):
            return self._value == other._value
        elif isinstance(other, bytes):
            return self._value == self._capitalize(other)
        return super().__eq__(other)

    def __lt__(self, other) -> bool:
        if isinstance(other, Flag):
            other_bytes = bytes(other)
        elif isinstance(other, bytes):
            other_bytes = self._capitalize(other)
        else:
            return NotImplemented
        if self.is_system and not other_bytes.startswith(b'\\'):
            return True
        elif not self.is_system and other_bytes.startswith(b'\\'):
            return False
        return bytes(self) < other_bytes

    def __hash__(self) -> int:
        return self._hash

    def __repr__(self) -> str:
        return '<{0} value={1!r}>'.format(type(self).__name__, bytes(self))

    def __bytes__(self) -> bytes:
        return self.value

    def __str__(self) -> str:
        return self.value.decode('ascii')

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> Tuple[Flag, memoryview]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        if buf:
            if buf[0] == 0x5c:
                atom, buf = Atom.parse(buf[1:], params)
                return cls(b'\\' + atom.value), buf
            else:
                atom, buf = Atom.parse(buf, params)
                return cls(atom.value), buf
        raise NotParseable(buf)


def get_system_flags() -> FrozenSet[Flag]:
    """Return the set of implemented system flags."""
    return frozenset({Seen, Recent, Deleted, Flagged, Answered, Draft})


#: The ``\\Seen`` system flag.
Seen = Flag(br'\Seen')

#: The ``\\Recent`` system flag.
Recent = Flag(br'\Recent')

#: The ``\\Deleted`` system flag.
Deleted = Flag(br'\Deleted')

#: The ``\\Flagged`` system flag.
Flagged = Flag(br'\Flagged')

#: The ``\\Answered`` system flag.
Answered = Flag(br'\Answered')

#: The ``\\Draft`` system flag.
Draft = Flag(br'\Draft')

#: The ``\\*`` special wildcard flag.
Wildcard = Flag(br'\*')




from __future__ import annotations

from bisect import bisect_right
from itertools import chain, groupby, islice
from typing import Any, Optional, Tuple, Dict, Set, MutableSet, AbstractSet, \
    FrozenSet, Iterable, List, Sequence
from weakref import WeakSet

from .flags import FlagOp, PermanentFlags, SessionFlags
from .interfaces.message import CachedMessage, FlagsKey
from .parsing.command import Command
from .parsing.primitives import ListP, Number
from .parsing.response import UntaggedResponse, ResponseBye
from .parsing.response.specials import ExistsResponse, RecentResponse, \
    ExpungeResponse, FetchResponse
from .parsing.specials import ObjectId, FetchAttribute, FetchValue, \
    Flag, SequenceSet

__all__ = ['SelectedSet', 'SynchronizedMessages', 'SelectedMailbox']

_flags_attr = FetchAttribute(b'FLAGS')
_uid_attr = FetchAttribute(b'UID')


class SelectedSet:
    """Maintains a weak set of :class:`SelectedMailbox` objects that exist for
    a mailbox, across all sessions. This is useful for assigning the
    ``\\Recent`` flag, as well as notifying other sessions about updates.

    """

    __slots__ = ['_set']

    def __init__(self) -> None:
        super().__init__()
        self._set: MutableSet[SelectedMailbox] = WeakSet()

    def add(self, selected: SelectedMailbox, *,
            replace: SelectedMailbox = None) -> None:
        """Add a new selected mailbox object to the set, which may then be
        returned by :meth:`.any_selected`.

        Args:
            selected: The new selected mailbox object.
            replace: An existing selected mailbox object that should be removed
                from the weak set.

        """
        if replace is not None:
            self._set.discard(replace)
        self._set.add(selected)

    @property
    def any_selected(self) -> Optional[SelectedMailbox]:
        """A single, random object in the set of selected mailbox objects.
        Selected mailbox object's marked :attr:`~SelectedMailbox.readonly`
        will not be chosen.

        """
        for selected in self._set:
            if not selected.readonly:
                return selected
        return None


class _Frozen:

    def __init__(self, selected: SelectedMailbox) -> None:
        super().__init__()
        messages = selected.messages
        session_flags = selected.session_flags
        self.is_deleted = selected._is_deleted
        self.uids = messages._uids.copy()
        self.seqs_cache = messages._seqs_cache.copy()
        self.flags = messages._flags_key_set.copy()
        self.recent = session_flags.recent_uids & self.uids
        self.sflags = frozenset(session_flags.flags.items())


class SynchronizedMessages:
    """Manages the message data that has been synchronized with the client."""

    def __init__(self) -> None:
        super().__init__()
        self._uids: Set[int] = set()
        self._sorted: List[int] = []
        self._seqs_cache: Dict[int, int] = {}
        self._cache: Dict[int, CachedMessage] = {}
        self._flags_key_map: Dict[int, FlagsKey] = {}
        self._flags_key_set: Set[FlagsKey] = set()
        self._pending_remove: Set[int] = set()

    @property
    def exists(self) -> int:
        """The total number of messages in the mailbox."""
        return len(self._uids)

    @property
    def max_uid(self) -> int:
        """The highest message UID value of the mailbox."""
        try:
            return self._sorted[-1]
        except IndexError:
            return 0

    def _update(self, messages: Iterable[CachedMessage]) -> None:
        lowest_idx: Optional[int] = None
        for msg in messages:
            msg_uid = msg.uid
            if msg_uid not in self._uids:
                self._uids.add(msg_uid)
                idx = bisect_right(self._sorted, msg_uid)
                if lowest_idx is None or lowest_idx > idx:
                    lowest_idx = idx
                self._sorted.insert(idx, msg_uid)
            self._cache[msg_uid] = msg
            new_flags_key = msg.flags_key
            old_flags_key = self._flags_key_map.get(msg_uid)
            if old_flags_key is not None:
                self._flags_key_set.discard(old_flags_key)
            self._flags_key_map[msg_uid] = new_flags_key
            self._flags_key_set.add(new_flags_key)
        if lowest_idx is not None:
            needs_reset = islice(self._sorted, lowest_idx, len(self._sorted))
            for seq, uid in enumerate(needs_reset, lowest_idx + 1):
                self._seqs_cache[uid] = seq

    def _remove(self, uids: Iterable[int], pending: bool) -> None:
        if pending:
            self._pending_remove.update(uids)
        else:
            any_removed = False
            for msg_uid in chain(uids, self._pending_remove):
                try:
                    self._uids.remove(msg_uid)
                except KeyError:
                    pass
                else:
                    flags_key = self._flags_key_map[msg_uid]
                    self._flags_key_set.remove(flags_key)
                    del self._flags_key_map[msg_uid]
                    del self._cache[msg_uid]
                    any_removed = True
            self._pending_remove.clear()
            if any_removed:
                self._sorted = sorted_uids = sorted(self._uids)
                self._seqs_cache = {uid: seq for seq, uid in
                                    enumerate(sorted_uids, 1)}

    def get(self, uid: int) -> Optional[CachedMessage]:
        """Return the given cached message.

        Args:
            uid: The message UID.

        """
        return self._cache.get(uid)

    def get_uids(self, seq_set: SequenceSet) -> Sequence[Tuple[int, int]]:
        """Return the message sequence numbers and their UIDs for the given
        sequence set.

        Args:
            seq_set: The message sequence set.

        """
        if seq_set.uid:
            all_uids = seq_set.flatten(self.max_uid) & self._uids
            return [(seq, uid) for seq, uid in enumerate(self._sorted, 1)
                    if uid in all_uids]
        else:
            all_seqs = seq_set.flatten(self.exists)
            return [(seq, uid) for seq, uid in enumerate(self._sorted, 1)
                    if seq in all_seqs]

    def get_all(self, seq_set: SequenceSet) \
            -> Sequence[Tuple[int, CachedMessage]]:
        """Return the cached messages, and their sequence numbers, for the
        given sequence set.

        Args:
            seq_set: The message sequence set.

        """
        if seq_set.uid:
            all_uids = seq_set.flatten(self.max_uid) & self._uids
            return [(seq, self._cache[uid])
                    for seq, uid in enumerate(self._sorted, 1)
                    if uid in all_uids]
        else:
            all_seqs = seq_set.flatten(self.exists)
            return [(seq, self._cache[uid])
                    for seq, uid in enumerate(self._sorted, 1)
                    if seq in all_seqs]


class SelectedMailbox:
    """Manages the updates to the current selected mailbox from other
    concurrent sessions.

    The current state of the selected mailbox will be written to this object by
    the backend implementation during each operation.  Then, when the operation
    completes, call :meth:`.fork` to make a fresh copy of the object and any
    untagged responses that should be added to the response.

    Args:
        mailbox_id: The globally unique identifier of the selected mailbox.
        readonly: Indicates the mailbox is selected as read-only.
        permanent_flags: The defined permanent flags for the mailbox.
        session_flags: Session-only flags for the mailbox.
        selected_set: The ``self`` object and subsequent forked copies will be
            kept in in this set.

    """

    __slots__ = ['_mailbox_id', '_readonly', '_permanent_flags',
                 '_session_flags', '_selected_set', '_kwargs', '_lookup',
                 '_mod_sequence', '_is_deleted', '_hide_expunged',
                 '_silenced_flags', '_silenced_sflags', '_prev', '_messages',
                 '__weakref__']

    def __init__(self, mailbox_id: ObjectId, readonly: bool,
                 permanent_flags: PermanentFlags,
                 session_flags: SessionFlags,
                 selected_set: SelectedSet = None,
                 lookup: Any = None, **kwargs: Any) -> None:
        super().__init__()
        self._mailbox_id = mailbox_id
        self._readonly = readonly
        self._permanent_flags = permanent_flags
        self._session_flags = session_flags
        self._selected_set = selected_set
        self._kwargs = kwargs
        self._lookup: Any = lookup
        self._mod_sequence = kwargs.get('_mod_sequence')
        self._is_deleted = False
        self._hide_expunged = False
        self._silenced_flags: Set[Tuple[int, FrozenSet[Flag]]] = set()
        self._silenced_sflags: Set[Tuple[int, FrozenSet[Flag]]] = set()
        self._prev: Optional[_Frozen] = kwargs.get('_prev')
        try:
            self._messages: SynchronizedMessages = kwargs['_messages']
        except KeyError:
            self._messages = SynchronizedMessages()
        if selected_set is not None:
            selected_set.add(self)

    @property
    def mailbox_id(self) -> ObjectId:
        """The selected mailbox object ID.

        See Also:
            :attr:`~pymap.interfaces.mailbox.MailboxInterface.mailbox_id`

        """
        return self._mailbox_id

    @property
    def lookup(self) -> Any:
        """The lookup value, if any, needed by backends that cannot lookup
        mailboxes by :attr:`.mailbox_id`. A typical lookup value might be the
        name of the mailbox.

        """
        return self._lookup

    @lookup.setter
    def lookup(self, lookup: Any) -> None:
        self._lookup = lookup

    @property
    def mod_sequence(self) -> Any:
        """The highest modification sequence of the mailbox."""
        return self._mod_sequence

    @mod_sequence.setter
    def mod_sequence(self, mod_sequence: Any) -> None:
        self._mod_sequence = mod_sequence

    @property
    def hide_expunged(self) -> bool:
        """If True, no untagged ``EXPUNGE`` responses will be generated, and
        message sequence numbers will not be adjusted, until the next
        :meth:`.fork`.

        """
        return self._hide_expunged

    @hide_expunged.setter
    def hide_expunged(self, hide_expunged: bool) -> None:
        self._hide_expunged = hide_expunged

    def add_updates(self, messages: Iterable[CachedMessage],
                    expunged: Iterable[int]) -> None:
        """Update the messages in the selected mailboxes. The ``messages``
        should include non-expunged messages in the mailbox that should be
        checked for updates. The ``expunged`` argument is the set of UIDs that
        have been expunged from the mailbox.

        In an optimized implementation, ``messages`` only includes new messages
        or messages with metadata updates.  This minimizes the comparison
        needed to determine what untagged responses are necessary. The
        :attr:`.mod_sequence` attribute may be used to support this
        optimization.

        If a backend implementation lacks the ability to determine the subset
        of messages that have been updated, it should instead use
        :meth:`.set_messages`.

        Args:
            messages: The cached message objects to add.
            expunged: The set of message UIDs that have been expunged.

        """
        self._messages._update(messages)
        self._messages._remove(expunged, self._hide_expunged)
        if not self._hide_expunged:
            self._session_flags.remove(expunged)

    def set_messages(self, messages: Sequence[CachedMessage]) -> None:
        """This is the non-optimized alternative to :meth:`.add_updates` for
        backend implementations that cannot detect their own updates and must
        instead compare the entire state of the mailbox.

        The ``messages`` list should contain the entire set of messages in the
        mailbox, ordered by UID. Any UID that previously existed and is not
        included in ``messages`` will be expunged.

        Args:
            messages: The entire set of cached message objects.

        """
        uids = {msg.uid for msg in messages}
        expunged = self._messages._uids - uids
        return self.add_updates(messages, expunged)

    @property
    def readonly(self) -> bool:
        """Indicates the mailbox is selected as read-only."""
        return self._readonly

    @property
    def messages(self) -> SynchronizedMessages:
        """The messages in the mailbox, as synchronized with the client."""
        return self._messages

    @property
    def permanent_flags(self) -> PermanentFlags:
        """The defined permanent flags for the mailbox."""
        return self._permanent_flags

    @property
    def session_flags(self) -> SessionFlags:
        """Session-only flags for the mailbox."""
        return self._session_flags

    def set_deleted(self) -> None:
        """Marks the selected mailbox as having been deleted."""
        self._is_deleted = True

    def silence(self, seq_set: SequenceSet, flag_set: AbstractSet[Flag],
                flag_op: FlagOp) -> None:
        """Runs the flags update against the cached flags, to prevent untagged
        FETCH responses unless other updates have occurred.

        For example, if a session adds ``\\Deleted`` and calls this method,
        the FETCH response will be silenced. But if another added ``\\Seen``
        at the same time, the FETCH response will be sent.

        Args:
            seq_set: The sequence set of messages.
            flag_set: The set of flags for the update operation.
            flag_op: The mode to change the flags.

        """
        session_flags = self.session_flags
        permanent_flag_set = self.permanent_flags & flag_set
        session_flag_set = session_flags & flag_set
        for seq, msg in self._messages.get_all(seq_set):
            msg_flags = msg.permanent_flags
            msg_sflags = session_flags.get(msg.uid)
            updated_flags = flag_op.apply(msg_flags, permanent_flag_set)
            updated_sflags = flag_op.apply(msg_sflags, session_flag_set)
            if msg_flags != updated_flags:
                self._silenced_flags.add((msg.uid, updated_flags))
            if msg_sflags != updated_sflags:
                self._silenced_sflags.add((msg.uid, updated_sflags))

    def fork(self, command: Command) \
            -> Tuple[SelectedMailbox, Iterable[UntaggedResponse]]:
        """Compares the state of the current object to that of the last fork,
        returning the untagged responses that reflect any changes. A new copy
        of the object is also returned, ready for the next command.

        Args:
            command: The command that was finished.

        """
        frozen = _Frozen(self)
        cls = type(self)
        copy = cls(self._mailbox_id, self._readonly, self._permanent_flags,
                   self._session_flags, self._selected_set, self._lookup,
                   _mod_sequence=self._mod_sequence,
                   _prev=frozen, _messages=self._messages)
        if self._prev is not None:
            with_uid: bool = getattr(command, 'uid', False)
            untagged = self._compare(self._prev, frozen, with_uid)
        else:
            untagged = []
        return copy, untagged

    def _compare(self, before: _Frozen, after: _Frozen,
                 with_uid: bool) -> Iterable[UntaggedResponse]:
        if after.is_deleted:
            yield ResponseBye(b'Selected mailbox no longer exists.')
            return
        cache = self._messages._cache
        session_flags = self._session_flags
        expunged_uids = before.uids - after.uids
        new_uids = after.uids - before.uids
        if not self._hide_expunged and expunged_uids:
            for uid in sorted(expunged_uids, reverse=True):
                yield ExpungeResponse(before.seqs_cache[uid])
        if new_uids:
            yield ExistsResponse(len(after.uids))
        if len(after.recent) != len(before.recent):
            yield RecentResponse(len(after.recent))
        new_recent = (after.recent - before.recent)
        new_flags = (after.flags - before.flags - self._silenced_flags)
        new_sflags = (after.sflags - before.sflags - self._silenced_sflags)
        fetch_uids = chain(new_recent,
                           (uid for uid, _ in new_flags),
                           (uid for uid, _ in new_sflags))
        for uid, _ in groupby(sorted(fetch_uids)):
            seq = after.seqs_cache[uid]
            msg_flags = cache[uid].get_flags(session_flags)
            fetch_data: List[FetchValue] = [
                FetchValue.of(_flags_attr, ListP(msg_flags, sort=True))]
            if with_uid:
                fetch_data.append(FetchValue.of(_uid_attr, Number(uid)))
            yield FetchResponse(seq, fetch_data)




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from contextlib import closing, asynccontextmanager, AsyncExitStack
from typing import Any, Type, Optional, Mapping, Dict, AsyncGenerator
from typing_extensions import Final

from pymap.context import connection_exit
from pymap.exceptions import InvalidAuth, ResponseError
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import IdentityInterface
from pymap.interfaces.session import SessionInterface
from pymap.plugin import Plugin
from pymapadmin.grpc.admin_pb2 import Result, SUCCESS, FAILURE
from pysasl import AuthenticationCredentials

from ..errors import get_unimplemented_error
from ..token import TokenCredentials

__all__ = ['handlers', 'BaseHandler', 'LoginHandler']

#: Registers new admin handler plugins.
handlers: Plugin[Type[BaseHandler]] = Plugin('pymap.admin.handlers')


class BaseHandler(metaclass=ABCMeta):
    """Base class for implementing admin request handlers.

    Args:
        backend: The backend in use by the system.
        admin_token: The admin token string that can authenticate any admin
            operation.

    """

    def __init__(self, backend: BackendInterface,
                 admin_token: Optional[bytes]) -> None:
        super().__init__()
        self.config: Final = backend.config
        self.login: Final = backend.login
        self.admin_token: Final = admin_token

    @abstractmethod
    def __mapping__(self) -> Dict[str, Any]:
        # Remove if IServable becomes public API in grpclib
        ...

    @asynccontextmanager
    async def catch_errors(self, command: str) -> AsyncGenerator[Result, None]:
        """Context manager to catch
        :class:`~pymap.exceptions.ResponseError` exceptions and include them in
        the response.

        Args:
            command: The admin command name.

        """
        response = b'. OK %b completed.' % command.encode('utf-8')
        result = Result(code=SUCCESS, response=response)
        async with AsyncExitStack() as stack:
            connection_exit.set(stack)
            try:
                yield result
            except NotImplementedError as exc:
                raise get_unimplemented_error() from exc
            except ResponseError as exc:
                result.code = FAILURE
                result.response = bytes(exc.get_response(b'.'))
                result.key = type(exc).__name__

    @asynccontextmanager
    async def login_as(self, metadata: Mapping[str, str], user: str) \
            -> AsyncGenerator[IdentityInterface, None]:
        """Context manager to login an identity object.

        Args:
            stream: The grpc request/response stream.
            user: The user to authorize as.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        try:
            creds: AuthenticationCredentials = TokenCredentials(
                metadata['auth-token'], self.admin_token, user)
        except (KeyError, ValueError) as exc:
            raise InvalidAuth() from exc
        yield await self.login.authenticate(creds)

    @asynccontextmanager
    async def with_session(self, identity: IdentityInterface) \
            -> AsyncGenerator[SessionInterface, None]:
        """Context manager to create a mail session for the identity.

        Args:
            identity: The authenticated user identity.

        Raises:
            :class:`~pymap.exceptions.InvalidAuth`

        """
        stack = connection_exit.get()
        session = await stack.enter_async_context(identity.new_session())
        stack.enter_context(closing(session))
        yield session




from __future__ import annotations

from abc import abstractmethod
from typing import TypeVar, Type, Optional, Tuple, Sequence
from typing_extensions import Protocol

from ..parsing.message import AppendMessage

__all__ = ['FilterValueT', 'FilterInterface',
           'FilterCompilerInterface', 'FilterSetInterface']

#: Type variable for the filter value representation.
FilterValueT = TypeVar('FilterValueT')


class FilterInterface(Protocol):
    """Protocol defining the interface for message filters. Filters may choose
    the mailbox or discard the message or modify its contents or metadata. The
    filter may also trigger external functionality, such as sending a vacation
    auto-response or a copy of the message to an SMTP endpoint.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    async def apply(self, sender: str, recipient: str, mailbox: str,
                    append_msg: AppendMessage) \
            -> Tuple[Optional[str], AppendMessage]:
        """Run the filter and return the mailbox where it should be appended,
        or None to discard, and the message to be appended, which is usually
        the same as ``append_msg``.

        Args:
            sender: The envelope sender of the message.
            recipient: The envelope recipient of the message.
            mailbox: The intended mailbox to append the message.
            append_msg: The message to be appended.

        Raises:
            :exc:`~pymap.exceptions.AppendFailure`

        """
        ...


class FilterCompilerInterface(Protocol[FilterValueT]):
    """Protocol for classes which can compile a filter value into an
    implementation.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def value_type(self) -> Type[FilterValueT]:
        """The filter value type."""
        ...

    @property
    @abstractmethod
    def filter_type(self) -> Type[FilterInterface]:
        """The filter implementation type."""
        ...

    @abstractmethod
    async def compile(self, value: FilterValueT) -> FilterInterface:
        """Compile the filter value and return the resulting implementation.

        Args:
            value: The filter value:

        """
        ...


class FilterSetInterface(Protocol[FilterValueT]):
    """Protocol defining the interface for accessing and managing the set of
    message filters currently active and available. This interface
    intentionally resembles that of a ManageSieve server.

    See Also:
        `RFC 5804 <https://tools.ietf.org/html/rfc5804>`_

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def compiler(self) -> FilterCompilerInterface[FilterValueT]:
        """Compiles filter values into an implementation."""
        ...

    @abstractmethod
    async def put(self, name: str, value: FilterValueT) -> None:
        """Add or update the named filter value.

        Args:
            name: The filter name.
            value: The filter value.

        """
        ...

    @abstractmethod
    async def delete(self, name: str) -> None:
        """Delete the named filter. If the named filter is active,
        :class:`ValueError` is raised.

        Args:
            name: The filter name.

        Raises:
            :class:`KeyError`, :class:`ValueError`

        """
        ...

    @abstractmethod
    async def rename(self, before_name: str, after_name: str) -> None:
        """Rename the filter, maintaining its active status. An exception is
        raised if ``before_name`` does not exist or ``after_name`` already
        exists.

        Args:
            before_name: The current filter name.
            after_name: The intended filter name.

        Raises:
            :class:`KeyError`

        """
        ...

    @abstractmethod
    async def clear_active(self) -> None:
        """Disable any active filters."""
        ...

    @abstractmethod
    async def set_active(self, name: str) -> None:
        """Set the named filter to be active.

        Args:
            name: The filter name.

        Raises:
            :class:`KeyError`

        """

    @abstractmethod
    async def get(self, name: str) -> FilterValueT:
        """Return the named filter value.

        Args:
            name: The filter name.

        Raises:
            :class:`KeyError`

        """
        ...

    @abstractmethod
    async def get_active(self) -> Optional[FilterValueT]:
        """Return the active filter value, if any."""
        ...

    @abstractmethod
    async def get_all(self) -> Tuple[Optional[str], Sequence[str]]:
        """Return the active filter name and a list of all filter names."""
        ...




from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import TypeVar, Tuple, Sequence, FrozenSet, Collection
from typing_extensions import Protocol

from ..bytes import Writeable
from ..flags import SessionFlags
from ..parsing.response.fetch import EnvelopeStructure, BodyStructure
from ..parsing.specials import Flag, ObjectId, FetchRequirement

__all__ = ['MessageT', 'MessageT_co', 'FlagsKey', 'CachedMessage',
           'MessageInterface', 'LoadedMessageInterface']

#: Type variable with an upper bound of :class:`MessageInterface`.
MessageT = TypeVar('MessageT', bound='MessageInterface')

#: Covariant type variable with an upper bound of :class:`MessageInterface`.
MessageT_co = TypeVar('MessageT_co', bound='MessageInterface', covariant=True)

#: Type alias for the value used as a key in set comparisons detecting flag
#: updates.
FlagsKey = Tuple[int, FrozenSet[Flag]]


class CachedMessage(Protocol):
    """Cached message metadata used to track state changes. Used to produce
    untagged FETCH responses when a message's flags have changed, or when a
    FETCH command requests metadata of an expunged message before its untagged
    EXPUNGE response has been sent.

    This is intended to be compatible with :class:`MessageInterface`, and
    should be implemented by the same classes in most cases.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def uid(self) -> int:
        """The message's unique identifier in the mailbox."""
        ...

    @property
    @abstractmethod
    def internal_date(self) -> datetime:
        """The message's internal date."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags for the message."""
        ...

    @property
    @abstractmethod
    def email_id(self) -> ObjectId:
        """The message's email object ID."""
        ...

    @property
    @abstractmethod
    def thread_id(self) -> ObjectId:
        """The message's thread object ID."""
        ...

    @property
    @abstractmethod
    def flags_key(self) -> FlagsKey:
        """Hashable value that represents the current flags of this
        message, used for detecting mailbox updates.

        """
        ...

    @abstractmethod
    def get_flags(self, session_flags: SessionFlags) -> FrozenSet[Flag]:
        """Get the full set of permanent and session flags for the message.

        Args:
            session_flags: The current session flags.

        """
        ...


class MessageInterface(Protocol):
    """Message data such as UID, permanent flags, and when the message was
    added to the system.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def uid(self) -> int:
        """The message's unique identifier in the mailbox."""
        ...

    @property
    @abstractmethod
    def expunged(self) -> bool:
        """True if this message has been expunged from the mailbox."""
        ...

    @property
    @abstractmethod
    def internal_date(self) -> datetime:
        """The message's internal date."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> FrozenSet[Flag]:
        """The permanent flags for the message."""
        ...

    @property
    @abstractmethod
    def email_id(self) -> ObjectId:
        """The message's email object ID, which can identify its content.

        See Also:
            `RFC 8474 5.1. <https://tools.ietf.org/html/rfc8474#section-5.1>`_

        """
        ...

    @property
    @abstractmethod
    def thread_id(self) -> ObjectId:
        """The message's thread object ID, which groups messages together.

        See Also:
            `RFC 8474 5.2. <https://tools.ietf.org/html/rfc8474#section-5.2>`_

        """
        ...

    @abstractmethod
    def get_flags(self, session_flags: SessionFlags) -> FrozenSet[Flag]:
        """Get the full set of permanent and session flags for the message.

        Args:
            session_flags: The current session flags.

        """
        ...

    @abstractmethod
    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessageInterface:
        """Loads the content of the message.

        Args:
            requirement: The data required from the message content.

        """
        ...


class LoadedMessageInterface(Protocol):
    """The loaded message content, which may include the header, the body,
    both, or neither, depending on the requirements.

    It is assumed that this object contains the entire content in-memory. As
    such, when multiple :class:`MessageInterface` objects are being processed,
    only one :class:`LoadedMessageInterface` should be in scope at a time.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def requirement(self) -> FetchRequirement:
        """The :class:`~pymap.parsing.specials.FetchRequirement` used to load
        the message content.

        """
        ...

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...

    @abstractmethod
    def get_header(self, name: bytes) -> Sequence[str]:
        """Get the values of a header from the message.

        Args:
            name: The name of the header.

        """
        ...

    @abstractmethod
    def get_headers(self, section: Sequence[int]) -> Writeable:
        """Get the headers from the message part.

        The ``section`` argument indexes a nested sub-part of the message. For
        example, ``[2, 3]`` would get the 2nd sub-part of the message and then
        index it for its 3rd sub-part.

        Args:
            section: Nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_message_headers(self, section: Sequence[int] = None,
                            subset: Collection[bytes] = None,
                            inverse: bool = False) -> Writeable:
        """Get the headers from the message or a ``message/rfc822`` sub-part of
        the message..

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            subset: Subset of headers to get.
            inverse: If ``subset`` is given, this flag will invert it so that
                the headers *not* in ``subset`` are returned.

        """
        ...

    @abstractmethod
    def get_message_text(self, section: Sequence[int] = None) -> Writeable:
        """Get the text of the message part, not including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_body(self, section: Sequence[int] = None,
                 binary: bool = False) -> Writeable:
        """Get the full body of the message part, including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            binary: True if the result has decoded any
                Content-Transfer-Encoding.

        """
        ...

    @abstractmethod
    def get_size(self, section: Sequence[int] = None) -> int:
        """Return the size of the message, in octets.

        Args:
            section: Optional nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_envelope_structure(self) -> EnvelopeStructure:
        """Build and return the envelope structure.

        See Also:
            `RFC 3501 2.3.5.
            <https://tools.ietf.org/html/rfc3501#section-2.3.5>`_

        """
        ...

    @abstractmethod
    def get_body_structure(self) -> BodyStructure:
        """Build and return the body structure.

        See Also:
            `RFC 3501 2.3.6
            <https://tools.ietf.org/html/rfc3501#section-2.3.6>`_

        """
        ...

    @abstractmethod
    def contains(self, value: bytes) -> bool:
        """Check the body of the message for a sub-string. This may be
        optimized to only search headers and ``text/*`` MIME parts.

        Args:
            value: The sub-string to find.

        """
        ...




from __future__ import annotations

import asyncio
import logging
from contextlib import closing
from typing import ClassVar, Callable, Awaitable, NoReturn

from aioredis import Redis, ConnectionClosedError  # type: ignore

from .keys import GlobalKeys, CleanupKeys, NamespaceKeys, ContentKeys, \
    MailboxKeys
from .scripts.cleanup import CleanupScripts

__all__ = ['CleanupTask', 'CleanupThread']

_log = logging.getLogger(__name__)
_scripts = CleanupScripts()


class CleanupTask:
    """Maintains a :class:`CleanupThread` for the duration of the process
    lifetime, restarting on failure.

    Args:
        connect_redis: Supplies a connected redis object.
        root: The root redis key.

    """

    #: The delay between redis reconnect attempts, on connection failure.
    connection_delay: ClassVar[float] = 5.0

    def __init__(self, connect_redis: Callable[[], Awaitable[Redis]],
                 global_keys: GlobalKeys) -> None:
        super().__init__()
        self._connect_redis = connect_redis
        self._global_keys = global_keys

    async def start(self) -> Awaitable:
        return self.run_forever()

    async def run_forever(self) -> NoReturn:
        """Run the cleanup loop indefinitely."""
        while True:
            try:
                with closing(await self._connect_redis()) as redis:
                    await CleanupThread(redis, self._global_keys).run()
            except (ConnectionClosedError, OSError):
                _log.warning('Redis connection failure', exc_info=True)
            await asyncio.sleep(self.connection_delay)


class CleanupThread:
    """Defines the logic for monitoring and executing cleanup of various
    entities.

    Args:
        redis: The redis connection object.
        global_keys: The global keys group.

    """

    namespace_ttl: ClassVar[int] = 0
    mailbox_ttl: ClassVar[int] = 600
    content_ttl: ClassVar[int] = 3600

    def __init__(self, redis: Redis, global_keys: GlobalKeys) -> None:
        super().__init__()
        self._redis = redis
        self._global_keys = global_keys
        self._keys = keys = CleanupKeys(global_keys)
        self._order = (keys.mailboxes, keys.namespaces, keys.contents)

    async def run(self) -> NoReturn:
        """Run the cleanup loop indefinitely.

        Raises:
            :class:`~aioredis.ConnectionClosedError`: The connection to redis
                was interrupted.

        """
        redis = self._redis
        while True:
            await redis.unwatch()
            cleanup_key, cleanup_val = await redis.blpop(
                *self._order, timeout=0)
            try:
                await asyncio.shield(self._run_one(cleanup_key, cleanup_val))
            except Exception:
                _log.warning('Cleanup failed: key=%s val=%s',
                             cleanup_key, cleanup_val, exc_info=True)
                raise

    async def _run_one(self, cleanup_key: bytes, cleanup_val: bytes) -> None:
        keys = self._keys
        if cleanup_key == keys.namespaces:
            namespace = cleanup_val
            await self._run_namespace(namespace)
        elif cleanup_key == keys.mailboxes:
            namespace, mailbox_id = cleanup_val.split(b'\x00', 1)
            await self._run_mailbox(namespace, mailbox_id)
        elif cleanup_key == keys.contents:
            namespace, email_id = cleanup_val.split(b'\x00', 1)
            await self._run_content(namespace, email_id)

    async def _run_namespace(self, namespace: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        await _scripts.namespace(self._redis, self._keys, ns_keys,
                                 ttl=self.namespace_ttl)

    async def _run_mailbox(self, namespace: bytes, mailbox_id: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        mbx_keys = MailboxKeys(ns_keys, mailbox_id)
        await _scripts.mailbox(self._redis, self._keys, mbx_keys,
                               ttl=self.mailbox_ttl)

    async def _run_content(self, namespace: bytes, email_id: bytes) -> None:
        ns_keys = NamespaceKeys(self._global_keys, namespace)
        ct_keys = ContentKeys(ns_keys, email_id)
        await _scripts.content(self._redis, ns_keys, ct_keys,
                               ttl=self.content_ttl)
